import socket
from functools import wraps
from .task import Task
from . import helpers
from .settings import HOST, PORT


class Leek(object):
    def task(self, f):
        @wraps(f)
        def _offload(*args, **kwargs):
            return push_task_to_queue(f, *args, **kwargs)
        f.offload = _offload
        return f


def push_task_to_queue(a_callable, *args, **kwargs):
    """Original API"""
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)    
    new_task = Task(a_callable, *args, **kwargs)
    new_task = helpers.save_task_to_db(new_task)  # returns with db_id
    sock.connect((HOST, PORT))
    sock.send(helpers.serielize(new_task))
    received = sock.recv(1024)
    sock.close()

    return received



class Task(object):
    def __init__(self, a_callable, *args, **kwargs):
        assert callable(a_callable)
        self.task_callable = a_callable
        self.args = args
        self.kwargs = kwargs
        self.db_id = None
    
    def run(self):
        self.task_callable(*self.args, **self.kwargs)



import datetime
import pickle
import base64

from . import models
from .task import Task


def unpack(pickled_task):
    new_task =  pickle.loads(base64.b64decode(pickled_task))
    assert isinstance(new_task,Task)
    return new_task


def serielize(task):
    return base64.b64encode(pickle.dumps(task))


def save_task_to_db(new_task):
    t = models.QueuedTasks()
    pickled_task = serielize(new_task)
    t = models.QueuedTasks(pickled_task=pickled_task)
    t.save()
    new_task.db_id = t.id
    return new_task

def save_task_failed(task,exception):
    t = models.FailedTasks(task_id=task.db_id, exception=str(exception))
    t.save()
    
    
def save_task_success(task):
    t = models.SuccessTasks(task_id=task.db_id)
    t.save()



from .worker import Worker


def start():
    global worker_thread
    worker_thread = Worker()


def put_task(task):
    return worker_thread.put_task_on_queue(task)


def stop():
    return worker_thread.stop_thread()


def waiting():
    return worker_thread.status_waiting()


def hanled():
    return worker_thread.status_handled()



import logging
import socketserver
import threading

from . import worker_manager


log = logging.getLogger(__name__)


Dcommands = {
    'waiting': worker_manager.waiting,
    'handled': worker_manager.hanled,
    'stop': worker_manager.stop
}


class TaskSocketServer(socketserver.BaseRequestHandler):
    def handle(self):
        try:
            data = self.request.recv(5000).strip()
            if data in Dcommands.keys():
                log.info('Got command: "{}"'.format(data))
                try:
                    worker_response = Dcommands[data]()
                    response = (True, worker_response.encode(),)
                    self.request.send(str(response).encode())
                except Exception as e:
                    log.exception("command failed")
                    response =  (False, "TaskServer Command: {}".format(e).encode(),)
                    self.request.send(str(response).encode())
            else:
                # assume a serialized task
                log.info('Got a task')
                try:
                    response = worker_manager.put_task(data)
                    self.request.send(str(response).encode())
                except Exception as e:
                    log.exception("failed to queue task")
                    response = (False, "TaskServer Put: {}".format(e).encode(),)
                    self.request.send(str(response).encode())

        except OSError as e:
            # in case of network error, just log
            log.exception("network error")



#  Based on the Worker class from Python in a nutshell, by Alex Martelli
import logging
import os
import sys
import threading
import queue


import django


from . import settings
from . import helpers


log = logging.getLogger(__name__)


class Worker(threading.Thread):
    def __init__(self):
        threading.Thread.__init__(self, name="django-leek")
        self.setDaemon(1)
        self.worker_queue = queue.Queue()
        self.tasks_counter = 0
        self.start()

    def put_task_on_queue(self, new_pickled_task):
        try:
            new_task = helpers.unpack(new_pickled_task)
            self.tasks_counter += 1
            self.worker_queue.put(new_task)
            return True, "sent"
        except Exception as e:
            return False, "Worker: %s"%str(e)

    def run_task(self, task):
        task.run()

    def stop_thread(self, timeout=None):
        """Stop the thread and wait for it to end."""
        self.worker_queue.put(None)
        self.join()

    def status_waiting(self):
        return self.worker_queue.qsize()

    def status_handled(self):
        # all, success & failes
        return self.tasks_counter

    def run(self):
        # the code until the while statement does NOT run atomicaly
        # a thread while loop cycle is atomic
        # thread safe locals: L = threading.local(), then L.foo="baz"
        django.setup()
        log.info('Worker Starts')
        done = False
        while not done:
            try:
                task = self.worker_queue.get()
                if task is None:
                    done = True
                    break
                
                log.info('running task...')
                self.run_task(task)
                helpers.save_task_success(task)
                log.info('...successfully')
            except Exception as e:
                log.exception("...task failed")
                helpers.save_task_failed(task, e)
                
        self.worker_queue = None
        log.info('Worker stopped, {} tasks handled.'.format(self.tasks_counter))
